#!/usr/bin/env python3
#
# This file is part of the GROMACS molecular simulation package.
#
# Copyright 2012- The GROMACS Authors
# and the project initiators Erik Lindahl, Berk Hess and David van der Spoel.
# Consult the AUTHORS/COPYING files and https://www.gromacs.org for details.
#
# GROMACS is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public License
# as published by the Free Software Foundation; either version 2.1
# of the License, or (at your option) any later version.
#
# GROMACS is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with GROMACS; if not, see
# https://www.gnu.org/licenses, or write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA.
#
# If you want to redistribute modifications to GROMACS, please
# consider that scientific software is very special. Version
# control is crucial - bugs must be traceable. We will be happy to
# consider code for inclusion in the official distribution, but
# derived work must not be called official GROMACS. Details are found
# in the README & COPYING files - if they are missing, get the
# official version at https://www.gromacs.org.
#
# To help us fund GROMACS development, we humbly ask that you cite
# the research papers on the package. Check out https://www.gromacs.org.

"""Script for generating spherical test configurations."""

import math
import random
import sys


def dot(a, b):
    return sum(x * y for (x, y) in zip(a, b))


def norm(a):
    return math.sqrt(dot(a, a))


def angle(a, b):
    return math.degrees(math.acos(dot(a, b) / (norm(a) * norm(b))))


def minangle(a, list):
    minangle = 180
    minindex = -1
    for index, x in enumerate(list):
        xangle = angle(a, x)
        if xangle < minangle:
            minangle = xangle
            minindex = index
    return (minangle, minindex)


def get_single_vec():
    while True:
        x = random.randint(-1000, 1000) / 1000.0
        y = random.randint(-1000, 1000) / 1000.0
        z = random.randint(-1000, 1000) / 1000.0
        pos = (x, y, z)
        dist = norm(pos)
        if dist <= 1.0 and dist > 0.25:
            return pos


def write_gro_title(fp, title, atomcount):
    fp.write(title + "\n")
    fp.write("%5d\n" % (atomcount))


def write_gro_atom(fp, resnr, resname, atomname, index, x):
    fp.write(
        "%5d%-5s%5s%5d%8.3f%8.3f%8.3f\n"
        % (resnr, resname, atomname, index, x[0], x[1], x[2])
    )


def write_gro_box(fp, box):
    fp.write("%10.5f%10.5f%10.5f\n" % box)


random.seed(1097)
center = (0, 0, 0)
cutoff = 20
possamples = 500
negsamples = 500
totsamples = 10000

sys.stderr.write("Generating reference points\n")
refpoints = []
refpoints.append((0, 0, -1))
refpoints.append((-0.5, 0.6, 0.1))
refpoints.append((-0.5, -0.5, 0.25))
while len(refpoints) < 30:
    pos = get_single_vec()
    if pos[0] > 0 and pos[1] > pos[0] and pos[2] > 0:
        refpoints.append(pos)

sys.stderr.write("Generating test points\n")
postestpoints = []
negtestpoints = []
hits = 0
samplecount = 0
while (
    samplecount < totsamples
    or len(postestpoints) < possamples
    or len(negtestpoints) < negsamples
):
    pos = get_single_vec()
    (pangle, index) = minangle(pos, refpoints)
    if pangle < cutoff:
        hits += 1
        if len(postestpoints) < possamples:
            postestpoints.append(pos)
    if pangle > cutoff:
        if len(negtestpoints) < negsamples:
            negtestpoints.append(pos)
    samplecount += 1

cfrac = float(hits) / samplecount
errest = math.sqrt((cfrac - cfrac * cfrac) / samplecount)
sys.stderr.write("Cutoff: %f angles\n" % (cutoff))
sys.stderr.write("Estimated covered fraction: %f +- %f\n" % (cfrac, errest))

debugfp = open("debug.txt", "w")
fp = sys.stdout
count = 1 + len(refpoints) + len(postestpoints) + len(negtestpoints)
write_gro_title(
    fp,
    "Spherical test case, cutoff %f, cfrac %f +- %f" % (cutoff, cfrac, errest),
    count,
)
n = 1
write_gro_atom(fp, 1, "C", "C", n, center)
n += 1
for i in range(len(refpoints)):
    write_gro_atom(fp, 2, "R", "R", n, refpoints[i])
    n += 1
for i in range(len(postestpoints)):
    write_gro_atom(fp, 3, "TP", "TP", n, postestpoints[i])
    x = postestpoints[i]
    xangle, index = minangle(x, refpoints)
    refx = refpoints[index]
    debugfp.write(
        "%3d%8.3f%8.3f%8.3f  %4.1f  %2d%8.3f%8.3f%8.3f\n"
        % (n - 1, x[0], x[1], x[2], xangle, index, refx[0], refx[1], refx[2])
    )
    n += 1
for i in range(len(negtestpoints)):
    write_gro_atom(fp, 4, "TN", "TN", n, negtestpoints[i])
    n += 1
write_gro_box(fp, (10, 10, 10))
debugfp.close()



#
# This file is part of the GROMACS molecular simulation package.
#
# Copyright 2019- The GROMACS Authors
# and the project initiators Erik Lindahl, Berk Hess and David van der Spoel.
# Consult the AUTHORS/COPYING files and https://www.gromacs.org for details.
#
# GROMACS is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public License
# as published by the Free Software Foundation; either version 2.1
# of the License, or (at your option) any later version.
#
# GROMACS is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with GROMACS; if not, see
# https://www.gnu.org/licenses, or write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA.
#
# If you want to redistribute modifications to GROMACS, please
# consider that scientific software is very special. Version
# control is crucial - bugs must be traceable. We will be happy to
# consider code for inclusion in the official distribution, but
# derived work must not be called official GROMACS. Details are found
# in the README & COPYING files - if they are missing, get the
# official version at https://www.gromacs.org.
#
# To help us fund GROMACS development, we humbly ask that you cite
# the research papers on the package. Check out https://www.gromacs.org.

"""
Provide command line operation.
"""

__all__ = ["commandline_operation"]

import collections.abc
import functools
import os
import pathlib
import shutil
import subprocess
import typing
from typing import Iterable
from typing import Union

import gmxapi as gmx
import gmxapi.operation
from gmxapi import exceptions
from gmxapi import logger as root_logger
from gmxapi.datamodel import NDArray
from gmxapi.exceptions import MissingImplementationError
from gmxapi.exceptions import UsageError
from gmxapi.operation import OutputCollectionDescription
from gmxapi.utility import config as _config

# Module-level logger
logger = root_logger.getChild("commandline")
logger.info("Importing {}".format(__name__))


@functools.lru_cache()
def cli_executable() -> pathlib.Path:
    """Report the installed GROMACS command line executable."""
    path = _config().get("gmx_executable", None)
    if path is not None:
        path = pathlib.Path(os.path.abspath(path))
        if path.is_file():
            return path
    raise exceptions.FeatureNotAvailableError("GROMACS installation unavailable.")


@functools.lru_cache()
def cli_bindir() -> pathlib.Path:
    """Report the installed GROMACS binary directory."""
    path = _config().get("gmx_bindir", None)
    if path is not None:
        path = pathlib.Path(os.path.abspath(path))
        if path.is_dir():
            return path
    raise exceptions.FeatureNotAvailableError("GROMACS installation unavailable.")


_seen = set()


# Create an Operation that consumes a list and a boolean to produce a string and an integer.
#
# Wrap the defined function using a decorator that
#    * strips the `output` parameter from the signature
#    * provides `output` publishing proxy to the inner function and
#    * produce a result with attributes for
#       * file: mapping of output flags to output filenames
#       * stdout: process STDOUT
#       * stderr: porcess STDERR
#       * returncode: integer return code of wrapped command
#
# Note that the existence of the 'file' output map is expressed here, but
# the keys of the map are not implicit or set by the wrapped function.
# For the map to be non-empty, it must be defined before the resulting helper
# function is called.
@gmx.function_wrapper(
    output={
        "directory": str,
        "returncode": int,
        "stderr": str,
        "stdout": str,
    }
)
def cli(
    command: NDArray, shell: bool, env: dict, output, stdin: str = "", _exist_ok=True
):
    """Execute a command line program in a subprocess.

    Configure an executable in a subprocess. Executes when run in an execution
    Context, as part of a work graph or via gmx.run().

    Subprocesses should be run in separate working directories. The directory name is
    the responsibility of the workflow manager and execution manager (and derived from
    the operation id, tied to fingerprinting information).

    Until gmxapi can provide some support for workflow checkpointing, it is unclear
    how to react to the presence of working directories from previous invocations.
    Starting with gmxapi 0.3.0, users may (optionally) set *_exist_ok=False* to
    force an error when an existing working directory for the task is found.
    Otherwise, the command is executed (as with previous gmxapi releases), and
    the command is responsible for determining how to react to the presence of
    prior working files or output, if necessary.

    Shell processing is not enabled, but can be considered for a future version.
    This means that shell expansions such as environment variables, globbing (`*`),
    and other special symbols (like `~` for home directory) are not available.
    This allows a simpler and more robust implementation, as well as a better
    ability to uniquely identify the effects of a command line operation. If you
    think this disallows important use cases, please let us know.

    Arguments:
         env: The environment variables map for the subprocess.
         _exist_ok: If ``True`` (default), it is not an error for the
             working directory to exist already.
         command: a tuple (or list) to be the subprocess arguments, including *executable*
         output: mapping of command line flags to output filename arguments
         shell: unused (provides forward-compatibility)
         stdin (str): String input to send to STDIN (terminal input) of the executable.

    Multi-line text sent to *stdin* should be joined into a single string
    (e.g. ``'\n'.join(list_of_strings) + '\n'``).
    If multiple strings are provided to *stdin*, gmxapi will assume an ensemble,
    and will run one operation for each provided string.

    Only string input (:py:func:str) to *stdin* is currently supported.
    If you have a use case that requires streaming input or binary input,
    please open an issue or contact the author(s).

    Arguments are iteratively added to the command line with standard Python
    iteration, so you should use a tuple or list even if you have only one parameter.
    I.e. If you provide a string with `arguments="asdf"` then it will be passed as
    `... "a" "s" "d" "f"`. To pass a single string argument, `arguments=("asdf")`
    or `arguments=["asdf"]`.

    Example:
        Execute a command named ``exe`` that takes a flagged option for file name
        (stored in a local Python variable `my_filename`) and an `origin` flag
        that uses the next three arguments to define a vector.

            >>> my_filename = "somefilename"
            >>> result = cli(('exe', '--origin', 1.0, 2.0, 3.0, '-f', my_filename), shell=False)
            >>> assert hasattr(result, 'file')
            >>> assert hasattr(result, 'stdout')
            >>> assert hasattr(result, 'stderr')
            >>> assert hasattr(result, 'returncode')

    Returns:
        A data structure with attributes for each of the results.

    Result object attributes:
        * file: the mapping of CLI flags to filename strings resulting from the `output` kwarg
        * returncode: return code of the subprocess.
        * stderr: A string mapping from process STDERR; it will be the
                    error output (if any) if the process failed.
        * stdout: A string mapping from process STDOUT.
    """
    # In the operation implementation, we expect the `shell` parameter to be intercepted by the
    # wrapper and set to False.
    if shell:
        raise exceptions.UsageError("Operation does not support shell processing.")

    if stdin == "":
        stdin = None

    if isinstance(command, (str, bytes)):
        command = [command]
    command = list([str(arg) for arg in command])

    executable = shutil.which(command[0])
    if executable is None:
        executable = shutil.which(command[0], path=str(cli_bindir()))
    if executable is None:
        raise exceptions.ValueError(
            '"{}" is not found or not executable.'.format(command[0])
        )
    command[0] = str(executable)

    # The SessionResources concept only exists in the form of the PublishingDataProxy provided as
    # *output* at run time.
    resource_manager: gmxapi.operation.SourceResource = typing.cast(
        gmxapi.operation.DataProxyBase, output
    )._resource_instance
    try:
        op_id = typing.cast(
            gmxapi.operation.ResourceManager, resource_manager
        ).operation_id
    except AttributeError as e:
        # Note that the SourceResource interface does not yet specify a means to get a canonical identifier
        # for the node in the work graph.
        raise MissingImplementationError(
            f"{resource_manager} does not provide required interface. gmxapi.commandline.cli() only "
            f"supports gmxapi.operation.ResourceManager."
        ) from e
    assert op_id

    ensemble_member = typing.cast(
        gmxapi.operation.DataProxyBase, output
    )._client_identifier
    uid = str(op_id)
    if ensemble_member:
        uid += f"_{ensemble_member}"
    assert uid not in _seen
    _seen.add(uid)

    # Note that `abspath` does not resolve symbolic links. (See `pathlib.Path.resolve`)
    # Some users may find it useful to preserve that abstraction.
    _cwd = os.path.abspath(uid)
    if os.path.exists(_cwd):
        message = f"Work directory {_cwd} already exists."
        if _exist_ok:
            logger.info(message)
        else:
            raise UsageError(message)
    os.makedirs(_cwd, exist_ok=_exist_ok)

    # TODO: (FR9) Can OS input/output filehandles be a responsibility of
    #  the code providing 'resources'?

    logger.debug("executing subprocess: %s", " ".join(str(word) for word in command))

    try:
        completed_process = subprocess.run(
            command,
            cwd=_cwd,
            env=env,
            shell=shell,
            input=stdin,
            check=True,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            encoding="utf-8",
            universal_newlines=True,
        )
        returncode = completed_process.returncode

        stdout = completed_process.stdout
        stderr = completed_process.stderr

    except subprocess.CalledProcessError as e:
        logger.info(f"Non-zero return status when calling {e.cmd} in {_cwd}")
        stdout = e.stdout
        stderr = e.stderr
        returncode = e.returncode

    if stderr is not None:
        if stdout is not None:
            logger.debug("STDOUT:")
            for line in stdout.split("\n"):
                logger.debug(line)
        else:
            logger.debug("STDOUT is empty")
        logger.debug("STDERR:")
        for line in stderr.split("\n"):
            logger.debug(line)
    else:
        logger.debug("STDERR is empty")

    # Publish outputs.
    output.directory = _cwd
    output.stdout = stdout
    output.stderr = stderr
    output.returncode = returncode


def _flatten_dict(mapping: dict):
    """Convert a dict of cli options to a sequence of strings.

    *mapping* values may be scalars or sequences. Non-string sequences will be converted to
    individual elements.

    Returns:
        Iterable of strings.
    """
    for key, value in mapping.items():
        yield str(key)
        if isinstance(value, collections.abc.Iterable) and not isinstance(
            value, (str, bytes)
        ):
            yield from [str(element) for element in value]
        else:
            yield value


@gmx.function_wrapper()
def filemap_to_flag_list(filemap: dict) -> list:
    """Convert a map of command line flags and filenames to a list of command line arguments.

    Used to map inputs and outputs of command line tools to and from gmxapi data handles.
    User provides mappings of flags and filenames so that gmxapi can construct an
    executable command line.

    Primary use case is implicit. commandline_operation() instantiates this operation based on
    user input, and sends the output to cli()

    Arguments:
        filemap: key-value map of command line flags and filename arguments

    Returns:
        list of strings and/or gmxapi data references
    """
    assert isinstance(filemap, dict)
    result = list(_flatten_dict(filemap))
    return result


def commandline_operation(
    executable=None,
    arguments=(),
    input_files: Union[dict, Iterable[dict]] = None,
    output_files: Union[dict, Iterable[dict]] = None,
    stdin: Union[str, Iterable[str]] = None,
    env: Union[dict, Iterable[dict]] = None,
    **kwargs,
):
    """Helper function to define a new operation that executes a subprocess in gmxapi data flow.

    Define a new Operation for a particular executable and input/output parameter set.
    Generate a chain of operations to process the named key word arguments and handle
    input/output data dependencies.

    Note that the operation will be executed in a subprocess in an automatically
    generated subdirectory. See below for more information.

    Arguments:
        arguments: list of positional arguments to insert at ``argv[1]``
        env: Optional replacement for the environment variables seen by the subprocess.
        executable: name of an executable on the path
        input_files: mapping of command-line flags to input file paths
        output_files: mapping of command-line flags to output file names
        stdin (str): String input to send to STDIN (terminal input) of the executable (optional).

    .. versionchanged:: 0.5.0
        Relative paths in *input_files* are considered relative to the current working directory,
        and are immediately converted to absolute paths.

    Multi-line text sent to *stdin* should be joined into a single string.
    E.g.::

        commandline_operation(..., stdin='\\n'.join(list_of_strings) + '\\n')

    If multiple strings are provided to *stdin*, gmxapi will assume an ensemble,
    and will run one operation for each provided string.

    Only string input (:py:func:`str`) to *stdin* is currently supported.
    If you have a use case that requires streaming input or binary input,
    please open an issue or contact the author(s).

    .. versionchanged:: 0.3.0
        output_files paths are converted to absolute paths at run time.

    If non-absolute paths are provided to *output_files*, paths are resolved relative to the
    working directory of the command instance (not relative to the working directory of the
    workflow script).

    By default, *executable* runs in a subprocess that inherits its environment
    and resources from the Python interpreter.
    (See https://docs.python.org/3/library/subprocess.html#subprocess.run)

    .. versionadded:: 0.3.1
        If specified, *env* replaces the default environment variables map seen by *executable* in the
        subprocess.

    .. versionchanged:: 0.4.1
        If unspecified, *env* defaults to a **filtered** copy of the current environment
        (with MPI-related environment variables removed).
        See :py:func:`gmxapi.runtime.filtered_mpi_environ()`.

    In addition to controlling environment variables used for user-input, it may be
    necessary to adjust the environment to prevent the subprocess from inheriting variables that it
    should not. This is particularly relevant if the Python script is launched with ``mpiexec`` and
    then ``commandline_wrapper`` is used to launch an MPI-aware executable that may try to manage
    the MPI context. (Reference :issue:`4421`)

    When overriding the environment variables, don't forget to include basic variables
    like PATH that are necessary for the executable to run. `os.getenv` can help.
    E.g. ``commandline_operation(..., env={'PATH': os.getenv('PATH'), ...})``

    See Also:
        :py:func:`gmxapi.runtime.filtered_mpi_environ()`.

    Output:
        The output node of the resulting operation handle contains

        * ``directory``: filesystem path that was used as the working directory for the subprocess
        * ``file``: the mapping of CLI flags to filename strings resulting from the ``output_files`` kwarg
        * ``returncode``: return code of the subprocess.
        * ``stderr``: A string mapping from process STDERR; it will be the
                      error output (if any) if the process failed.
        * ``stdout``: A string mapping from process STDOUT.

    .. versionchanged:: 0.3
        Subprocesses run in directories managed by gmxapi.
    .. versionadded:: 0.3
        The *directory* output.

    Working directory names are details of the gmxapi implementation; the naming scheme is not
    yet specified by the API, but is intended to be related to the operation ID.

    Note that re-executing a gmxapi script in the same directory will cause
    commands to be executed again in the same directories. If this presents a
    risk of data corruption (or just wasted compute cycles), you may include
    the key word argument ``_exist_ok=False`` to force an error. Please consider
    contacting the developers through any of the various GROMACS community
    channels to further discuss your use case.
    """
    import gmxapi.runtime

    # Implementation details: When used in a script, this function returns an
    # instance of an operation. However, because of the dynamic specification of
    # inputs and outputs, each invocation may have the overhead of defining new
    # types to express the data flow topology, regardless of the executable.
    # If this overhead is problematic, consider exposing the intermediate step
    # at which the Operation is fully specified to facilitate reuse.

    ##
    # 1. Define a new operation with outputs from `cli()` plus `file` from `output_files`

    # output_files is essentially passed through, but we need assurance that results
    # will not be published until the rest of the operation has run (i.e. the cli() executable.)

    # Warning: decorating a local function like this is counter to the notion of Operations
    # as portable (importable, serializable/deserializable). The big picture here needs
    # some more consideration.
    # TODO(3139): Distinguish portable Operations from relocatable Futures.
    # There is nothing antithetical about objects implementing gmxapi data interfaces
    # that are only resolvable by a certain Context as long as that Context can convey
    # the results to another Context upon request. Re-instantiating Operations is
    # only one way of relocating Futures. In this case, though, the dynamic creation of
    # merged_ops doesn't seem right, and commandline_operation should probably be
    # a proper Operation.
    @gmx.function_wrapper(
        output={
            "directory": str,
            "file": dict,
            "returncode": int,
            "stderr": str,
            "stdout": str,
        }
    )
    def merged_ops(
        stdout: str = None,
        stderr: str = None,
        returncode: int = None,
        file: dict = None,
        directory: str = None,
        output: OutputCollectionDescription = None,
    ):
        assert stdout is not None
        assert stderr is not None
        assert returncode is not None
        assert file is not None
        assert output is not None
        output.returncode = returncode
        output.stdout = stdout
        output.stderr = stderr
        output.directory = directory
        file_map = file.copy()
        for key, value in file_map.items():
            if not os.path.isabs(value):
                value = os.path.abspath(os.path.join(directory, value))
            file_map[key] = value
        if returncode == 0:
            output.file = file_map
        else:
            output.file = {}

    ##
    # 2. Prepare data flow.

    if input_files is None:
        input_files = {}
    normalized_input_files = {}
    for key, value in input_files.items():
        if isinstance(value, (str, bytes, pathlib.Path)):
            normalized_input_files[key] = os.path.abspath(value)
        elif isinstance(value, Iterable):
            # Recurse only one layer into sequence values
            normalized_input_files[key] = list()
            for element in value:
                if isinstance(element, (str, bytes, pathlib.Path)):
                    element = os.path.abspath(element)
                normalized_input_files[key].append(element)
        else:
            normalized_input_files[key] = value
    input_files = normalized_input_files
    if output_files is None:
        output_files = {}
    try:
        executable = str(executable)
    except Exception as e:
        raise gmxapi.exceptions.TypeError(
            "gmxapi typing currently requires paths and names to be strings. *executable* argument is "
            f"{type(executable)}."
        )
    if isinstance(arguments, (str, bytes)):
        arguments = [arguments]
    input_options = filemap_to_flag_list(input_files).output.data
    output_options = filemap_to_flag_list(output_files).output.data
    command = gmx.concatenate_lists(
        [[executable], arguments, input_options, output_options]
    )
    shell = gmx.make_constant(False)

    if env is None:
        env = gmx.function_wrapper(allow_duplicate=True)(
            gmxapi.runtime.filtered_mpi_environ
        )().output.data
    cli_args = {"command": command, "shell": shell, "env": env}
    cli_args.update(**kwargs)
    if stdin is not None:
        cli_args["stdin"] = stdin

    ##
    # 3. Merge operations
    #
    # Note: Without a `label` argument, repeated calls to cli(**cli_args) should
    # produce references to the same unique resource. Creating this handle
    # separately should not be necessary, but we've got a way to go until we have the
    # fingerprinting and Context resource management we need for that.
    # TODO: ``label`` kwarg
    # TODO: input fingerprinting
    cli_result = cli(**cli_args)
    merged_result = merged_ops(
        stdout=cli_result.output.stdout,
        stderr=cli_result.output.stderr,
        returncode=cli_result.output.returncode,
        directory=cli_result.output.directory,
        file=output_files,
        **kwargs,
    )

    # Return an object with an OutputCollection granting access to outputs of
    # cli() and of output_files (as "file")
    return merged_result



#
# This file is part of the GROMACS molecular simulation package.
#
# Copyright 2019- The GROMACS Authors
# and the project initiators Erik Lindahl, Berk Hess and David van der Spoel.
# Consult the AUTHORS/COPYING files and https://www.gromacs.org for details.
#
# GROMACS is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public License
# as published by the Free Software Foundation; either version 2.1
# of the License, or (at your option) any later version.
#
# GROMACS is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with GROMACS; if not, see
# https://www.gnu.org/licenses, or write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA.
#
# If you want to redistribute modifications to GROMACS, please
# consider that scientific software is very special. Version
# control is crucial - bugs must be traceable. We will be happy to
# consider code for inclusion in the official distribution, but
# derived work must not be called official GROMACS. Details are found
# in the README & COPYING files - if they are missing, get the
# official version at https://www.gromacs.org.
#
# To help us fund GROMACS development, we humbly ask that you cite
# the research papers on the package. Check out https://www.gromacs.org.

"""Extra support for type hinting and generics.

Provides additional support for annotation and static type checking.
Extends the specifications of the abstract base classes in gmxapi.abc.
"""

from typing import Tuple, TypeVar, Generic, NewType, Type

import gmxapi.abc

# Use SourceTypeVar and ResultTypeVar for static type hints, annotations, and as a parameter to generics.
# Use valid_source_types and valid_result_types for run-time type checking.
ResultTypeVar = TypeVar(
    "ResultTypeVar", *(str, bool, int, float, dict, gmxapi.abc.NDArray)
)
valid_result_types: Tuple[type] = ResultTypeVar.__constraints__

SourceTypeVar = TypeVar(
    "SourceTypeVar", *(str, bool, int, float, dict, gmxapi.abc.NDArray)
)
valid_source_types: Tuple[type] = SourceTypeVar.__constraints__

# Place holder for type annotations of Context objects.
# TODO: Expand to support some static type checking.
_Context = TypeVar("_Context", bound=gmxapi.abc.Context)
# Type variable that binds to subclasses of the forward referenced OperationImplentation ABC.
_Op = TypeVar("_Op", bound=gmxapi.abc.OperationImplementation)

# Note: We could make many types generic in SourceContext and TargetContext and
# compose functionality into instances of a helpful base class with the option of
# subclassing the generic base with concrete types. If we want to reserve the word
# "Context" for an execution context, then we can say that a SourceResource or
# TargetEnvironment can each be either a Context or an OperationExecutionDirector.
#
# We can't
# generally add type-hinting details in terms of a specialized generic, but we
# can require a particular version of a generic and we can use the inferred
# type parameter value in limited cases, including to fully specify other generics.
# For instance, we can't assert that a generic with a particular parameter has
# additional attributes, but we can overload a function that produces an object
# with those attributes. More generally, using some sort of identity function like
# that may be the solution to the type erasure of a generic abstract base class.
# For static type checking of dynamically defined subclasses, we can use a decorator
# to reduce boiler plate if the type checker does a decent job of resolving closures.
# Ironically, there may be cases where we want to *avoid* annotating the return
# value to allow the static source analyzer to infer the dynamically typed return value.


class Future(Generic[ResultTypeVar]):
    """Generic result data."""

    def dtype(self) -> Type[ResultTypeVar]: ...

    def result(self) -> ResultTypeVar: ...


class OperationReference(gmxapi.abc.OperationReference, Generic[_Op]):
    """Object with an OperationReference interface.

    Generic version of AbstractOperationReference, parameterized by operation
    implementation.
    """


class OperationDirector(Generic[_Op, _Context]):
    """Generic abstract operation director.

    An operation director is instantiated for a specific operation and context
    by a dispatching factory to add a computational element to the work managed
    by the context.

    Note:
        This is a generic class, as defined with the Python `typing` module.
        If used as a base class, re-expression of the TypeVar parameters in class
        subscripts will cause the derived class to be generic unless regular
        classes (non-TypeVar) are given. Omission of the subscripts causes `Any`
        to be bound, which also results in a non-generic subclass.
    """


class OperationImplementation(Generic[_Op], gmxapi.abc.OperationImplementation):
    """Essential interface of an Operation implementation.

    Describe the essential features of an Operation that can be registered with
    gmxapi to support building and executing work graphs in gmxapi compatible
    execution contexts.
    """

    # The executable part of an operation consumes a distinct resource type.
    # The resource type may be opaque, because it is created through a factory
    # and only used in passing to a function object.
    ResourceType = NewType("ResourceType", object)

    def name(self) -> str:
        """The name of the operation.

        Generally, this corresponds to a callable attribute of a Python module
        (named by namespace()) that acts as a factory for new operation instances.
        It is also used by Context implementations to locate code supporting
        the operation.
        """
        ...

    def namespace(self) -> str:
        """The namespace of the operation.

        Generally, the namespace corresponds to a Python module importable in
        the execution environment.
        """

    # Note that this indicates that the ABC requires subclasses to provide a generic function,
    # which _is_ the factory behavior we are trying to specify.
    # TODO: Confirm that this type-checks correctly.
    # We may need to look more at how the type checking is implemented to see how
    # to do this right.
    def director(self, context: _Context) -> OperationDirector[_Op, _Context]:
        """Factory to get an OperationDirector appropriate for the context."""
        ...



#
# This file is part of the GROMACS molecular simulation package.
#
# Copyright 2019- The GROMACS Authors
# and the project initiators Erik Lindahl, Berk Hess and David van der Spoel.
# Consult the AUTHORS/COPYING files and https://www.gromacs.org for details.
#
# GROMACS is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public License
# as published by the Free Software Foundation; either version 2.1
# of the License, or (at your option) any later version.
#
# GROMACS is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with GROMACS; if not, see
# https://www.gnu.org/licenses, or write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA.
#
# If you want to redistribute modifications to GROMACS, please
# consider that scientific software is very special. Version
# control is crucial - bugs must be traceable. We will be happy to
# consider code for inclusion in the official distribution, but
# derived work must not be called official GROMACS. Details are found
# in the README & COPYING files - if they are missing, get the
# official version at https://www.gromacs.org.
#
# To help us fund GROMACS development, we humbly ask that you cite
# the research papers on the package. Check out https://www.gromacs.org.

"""Test gromacs.mdrun operation.

Factory produces deferred execution operation.

TODO: Factory expects input for conformation, topology, simulation parameters, simulation state.

TODO: Factory accepts additional keyword input to indicate binding
 to the "potential" interface.
"""

import logging
import os
import pathlib

import pytest

import gmxapi as gmx

from gmxapi.utility import join_path
from gmxapi.version import api_is_at_least

gmx.logger.setLevel(logging.WARNING)

try:
    from mpi4py import MPI

    rank_number = MPI.COMM_WORLD.Get_rank()
    comm_size = MPI.COMM_WORLD.Get_size()
except ImportError:
    rank_number = 0
    comm_size = 1
    rank_tag = ""
    MPI = None
else:
    rank_tag = "rank{}:".format(rank_number)


# Use this formatter to improve the caplog log records.
# formatter = logging.Formatter(rank_tag + '%(name)s:%(levelname)s: %(message)s')

# For additional console logging, create and attach a stream handler.
# For example:
#    ch = logging.StreamHandler()
#    ch.setFormatter(formatter)
#    logging.getLogger().addHandler(ch)


@pytest.mark.usefixtures("cleandir")
def test_run_from_tpr(spc_water_box, mdrun_kwargs):
    assert os.path.exists(spc_water_box)

    assert gmx.version.has_feature("mpi_comm_integration")

    md = gmx.mdrun(spc_water_box, runtime_args=mdrun_kwargs)
    md.run()

    published_trajectory = md.output.trajectory.result()
    expected_trajectory = join_path(
        md.output.directory, "traj.trr"
    ).output.path.result()
    assert os.path.exists(published_trajectory)
    assert published_trajectory == expected_trajectory

    # Look for some expected output text to make sure we are capturing
    # status messages from libgromacs as intended. Ref #4541
    stderr = md.output.stderr.result()
    assert os.path.exists(stderr)
    with open(stderr, "r") as fh:
        assert "starting mdrun" in fh.read()


@pytest.mark.skipif(
    not api_is_at_least(0, 4, 1), reason="runtime_args was refined for gmxapi 0.4.1."
)
def test_runtime_args(mdrun_kwargs):
    import gmxapi._gmxapi as core

    # Check the C++ bindings for our basic set of options.
    mdargs = core.MDArgs()
    mdargs.set(mdrun_kwargs)

    # Before gmxapi 0.4.1, a non-string value was not handled properly.
    mdrun_kwargs["-rdd"] = 2.5
    mdargs = core.MDArgs()
    mdargs.set(mdrun_kwargs)

    # Test non-scalar argument value.
    mdrun_kwargs["-dd"] = (1, 1, 1)
    mdargs = core.MDArgs()
    mdargs.set(mdrun_kwargs)

    # Test flag-like argument with no value.
    mdrun_kwargs["-noappend"] = None
    mdargs.set(mdrun_kwargs)
    assert all(arg != "" for arg in mdargs.get_args())

    argstring = " ".join(mdargs.get_args())
    assert "-dd 1 1 1" in argstring


@pytest.mark.usefixtures("cleandir")
def test_mdrun_runtime_args(spc_water_box, caplog, mdrun_kwargs):
    """Test that *runtime_args* is respected.

    When an ensemble is possible, confirm that an array of inputs
    causes each rank to produce a unique simulation with unique output.
    """
    assert api_is_at_least(0, 3)
    with caplog.at_level(logging.DEBUG):
        with caplog.at_level(logging.WARNING, "gmxapi"), caplog.at_level(
            logging.DEBUG, "gmxapi.mdrun"
        ), caplog.at_level(logging.DEBUG, "gmxapi.modify_input"), caplog.at_level(
            logging.DEBUG, "gmxapi.read_tpr"
        ), caplog.at_level(
            logging.DEBUG, "gmxapi.simulation"
        ):
            tpr = gmx.read_tpr([spc_water_box] * comm_size)
            # Note: It was originally assumed that each process (each rank) would be executing
            # with the same arguments at the user level and that ensemble differences would
            # happen within the framework. This is not enforced, but nor does it present a
            # particular problem. Nevertheless, pending further design conversations,
            # we will test the normative usage and refrain from doing fancy tricks with
            # rank-based input generation.
            traj_name = "unusualtrajectoryname.trr"
            runtime_args = {"-o": traj_name}
            runtime_args.update(mdrun_kwargs)
            md = gmx.mdrun(tpr, runtime_args=runtime_args)
            md.run()

            if comm_size == 1:
                output = md.output.trajectory.result()
            else:
                assert comm_size > 1
                output = md.output.trajectory.result()[rank_number]
            assert str(output).endswith(traj_name)
            assert os.path.exists(output)
            if comm_size > 1:
                assert (
                    md.output.trajectory.result()[0] != md.output.trajectory.result()[1]
                )


@pytest.mark.withmpi_only
@pytest.mark.usefixtures("cleandir")
def test_mdrun_parallel_runtime_args(spc_water_box, mdrun_kwargs):
    """Test that array input of *runtime_args* is respected.

    Confirm proper resolution of ensemble data flow.
    * Array of runtime_args causes broadcast of singular input.
    * Array of input and array of runtime_args are automatically mapped to parallel data flow.
    """
    assert comm_size > 1
    assert api_is_at_least(0, 3)

    ensemble_size = 2

    tpr = gmx.read_tpr(spc_water_box)
    output_files = [f"traj{i}.trr" for i in range(ensemble_size)]
    # Take care not to reference the same dictionary object.
    runtime_args = list(mdrun_kwargs.copy() for _ in range(ensemble_size))
    for i, name in enumerate(output_files):
        runtime_args[i].update({"-o": name})
    md = gmx.mdrun(tpr, runtime_args=runtime_args)
    md.run()
    # TODO: Update the message to make more sense with comm_size > ensemble_size
    logging.getLogger().debug(f"testing rank {rank_number} with comm size {comm_size}.")
    for _rank in range(ensemble_size):
        output = md.output.trajectory.result()[_rank]
        assert str(output).endswith(f"traj{_rank}.trr")
        assert os.path.exists(output)
        assert md.output.trajectory.result()[0] != md.output.trajectory.result()[1]

    tpr = gmx.read_tpr([spc_water_box] * comm_size)
    output_files = [f"traj{i}.trr" for i in range(comm_size)]
    runtime_args = list(mdrun_kwargs.copy() for i in range(comm_size))
    for i, name in enumerate(output_files):
        runtime_args[i].update({"-o": name})
    md = gmx.mdrun(tpr, runtime_args=runtime_args)
    md.run()
    logging.getLogger().debug(f"testing rank {rank_number} with comm size {comm_size}.")
    for _rank in range(comm_size):
        output = md.output.trajectory.result()[_rank]
        assert str(output).endswith(f"traj{_rank}.trr")
        assert os.path.exists(output)
        assert md.output.trajectory.result()[0] != md.output.trajectory.result()[1]


@pytest.mark.usefixtures("cleandir")
def test_extend_simulation_via_checkpoint(spc_water_box, mdrun_kwargs, caplog):
    assert os.path.exists(spc_water_box)
    assert api_is_at_least(0, 3)

    with caplog.at_level(logging.DEBUG):
        with caplog.at_level(logging.WARNING, "gmxapi"), caplog.at_level(
            logging.DEBUG, "gmxapi.mdrun"
        ), caplog.at_level(logging.DEBUG, "gmxapi.modify_input"), caplog.at_level(
            logging.DEBUG, "gmxapi.read_tpr"
        ), caplog.at_level(
            logging.DEBUG, "gmxapi.simulation"
        ):
            tpr = gmx.read_tpr(spc_water_box)
            parameters = dict({"nsteps": 2, "nstxout": 2})
            input1 = gmx.modify_input(tpr, parameters=parameters)
            runtime_args = {"-cpo": "continuation.cpt"}
            runtime_args.update(mdrun_kwargs)
            md1 = gmx.mdrun(input1, runtime_args=runtime_args)

            input2 = gmx.modify_input(tpr, parameters={"nsteps": 4, "nstxout": 2})
            runtime_args = {
                "-cpi": md1.output.checkpoint,
                "-cpo": "state.cpt",
                "-noappend": None,
            }
            runtime_args.update(mdrun_kwargs)
            md2 = gmx.mdrun(input2, runtime_args=runtime_args)
            md2.run()

            # Check for issue #4795 regression
            output_dir1 = pathlib.Path(md1.output.directory.result()).resolve()
            output_dir2 = pathlib.Path(md2.output.directory.result()).resolve()
            assert output_dir1.exists()
            assert output_dir2.exists()
            assert output_dir2 != output_dir1

            # By inspection of the output, we can see that the second trajectory has continued
            # from the checkpoint, but we cannot programmatically confirm it at this point.
            # TODO: Check more rigorously when we can read trajectory files.


@pytest.mark.withmpi_only
@pytest.mark.usefixtures("cleandir")
def test_run_trivial_ensemble(spc_water_box, caplog, mdrun_kwargs):
    with caplog.at_level(logging.DEBUG):
        with caplog.at_level(logging.WARNING, "gmxapi"), caplog.at_level(
            logging.DEBUG, "gmxapi.mdrun"
        ), caplog.at_level(logging.DEBUG, "gmxapi.modify_input"), caplog.at_level(
            logging.DEBUG, "gmxapi.read_tpr"
        ), caplog.at_level(
            logging.DEBUG, "gmxapi.simulation"
        ):
            tpr_filename = spc_water_box
            ensemble_width = 2
            simulation_input = gmx.read_tpr([tpr_filename] * ensemble_width)
            assert simulation_input.output.ensemble_width == ensemble_width
            assert (
                len(simulation_input.output._simulation_input.result())
                == ensemble_width
            )
            md = gmx.mdrun(simulation_input, runtime_args=mdrun_kwargs)
            assert md.output.ensemble_width == ensemble_width
            md.run()

            output_directory = md.output.directory.result()
            logging.info("output_directory result: {}".format(str(output_directory)))
            assert len(output_directory) == ensemble_width

            # Note that the 'cleandir' test fixture will clean up the output directory on
            # other ranks, so only check the current rank. Generally, our behavior
            # is undefined if the client removes the working directory while the job
            # is in progress. We can consider adding some sort of synchronization at
            # the end of the job if running in temporary directories becomes an
            # important use case outside of testing.
            assert output_directory[0] != output_directory[1]
            for _rank in range(ensemble_width):
                assert os.path.exists(output_directory[_rank])
                assert os.path.exists(md.output.trajectory.result()[_rank])


@pytest.mark.usefixtures("cleandir")
def test_run_from_read_tpr_op(spc_water_box, caplog, mdrun_kwargs):
    with caplog.at_level(logging.DEBUG):
        with caplog.at_level(logging.DEBUG, "gmxapi"):
            simulation_input = gmx.read_tpr(spc_water_box)
            md = gmx.mdrun(input=simulation_input, runtime_args=mdrun_kwargs)

            md.run()
            assert os.path.exists(md.output.trajectory.result())


@pytest.mark.usefixtures("cleandir")
def test_run_from_modify_input_op(spc_water_box, caplog, mdrun_kwargs):
    with caplog.at_level(logging.DEBUG):
        simulation_input = gmx.read_tpr(spc_water_box)
        modified_input = gmx.modify_input(
            input=simulation_input, parameters={"nsteps": 4}
        )
        md = gmx.mdrun(input=modified_input, runtime_args=mdrun_kwargs)

        md.run()



#
# This file is part of the GROMACS molecular simulation package.
#
# Copyright 2019- The GROMACS Authors
# and the project initiators Erik Lindahl, Berk Hess and David van der Spoel.
# Consult the AUTHORS/COPYING files and https://www.gromacs.org for details.
#
# GROMACS is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public License
# as published by the Free Software Foundation; either version 2.1
# of the License, or (at your option) any later version.
#
# GROMACS is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with GROMACS; if not, see
# https://www.gnu.org/licenses, or write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA.
#
# If you want to redistribute modifications to GROMACS, please
# consider that scientific software is very special. Version
# control is crucial - bugs must be traceable. We will be happy to
# consider code for inclusion in the official distribution, but
# derived work must not be called official GROMACS. Details are found
# in the README & COPYING files - if they are missing, get the
# official version at https://www.gromacs.org.
#
# To help us fund GROMACS development, we humbly ask that you cite
# the research papers on the package. Check out https://www.gromacs.org.

"""
gmxapi version and release information.

The ``gmxapi.__version__`` attribute contains a :pep:`version string <440>`.
The more general way to access the package version is with the
:py:mod:`pkg_resources <https://setuptools.readthedocs.io/en/latest/pkg_resources.html>` module::

    pkg_resources.get_distribution('gmxapi').version

`gmxapi.version` module functions `api_is_at_least()` and `has_feature()`
support additional convenience and introspection.

.. versionchanged:: 0.2

    This module no longer provides public data attributes.
    Instead, use the module functions or
    :py:mod:`packaging.version <https://packaging.pypa.io/en/latest/version/>`.

.. seealso::

    Consider https://packaging.pypa.io/en/latest/version/ for programmatic
    handling of the version string. For example::

        import pkg_resources
        from packaging.version import parse
        gmxapi_version = pkg_resources.get_distribution('gmxapi').version
        if parse(gmxapi_version).is_prerelease:
            print('The early bird gets the worm.')

"""
import warnings

from .exceptions import FeatureNotAvailableError

# TODO(#3851): Version management policy and procedures.
_major = 0
_minor = 5
_micro = 0
_suffix = "a1"

# Reference https://www.python.org/dev/peps/pep-0440/ and
# https://packaging.pypa.io/en/latest/version/ for
# versioning semantics and rules.
__version__ = "{major}.{minor}.{micro}{suffix}".format(
    major=_major, minor=_minor, micro=_micro, suffix=_suffix
)

# Named features describe functionality or behavior introduced since the last
# major release, and should be described in gmxapi documentation or issue
# tracking system. Note that, as features become part of the specification,
# conditionals depending on them should be phased out of the package source. At
# major releases, the named feature list should be reset to empty. Optionally,
# we could raise a DeprecationWarning for calls to has_feature() for features
# that have become part of the specification, at least for a few minor release or
# a few years, to avoid introducing errors to client code.
#
# Bugs and bug fixes may be indicated with names consisting of tracked issue URLs.

# Named features for gmxapi 0.x (pre-1.0 versions).
_named_features_0 = [[]] * (_minor + 1)

# Features added since the initial gmxapi prototype, targeted for version 0.1.
# Functional requirements were described in issues #2045 and #2893. See also
# https://gitlab.com/gromacs/gromacs/-/blob/release-2020/python_packaging/roadmap.rst
_named_features_0[0] = ["fr1", "fr3", "fr7", "fr15"]
# Features named since the finalization of the 0.1 specification with GROMACS 2020.
_named_features_0[1] = []

_named_features_0[2] = [
    "container_futures",
    "mdrun_checkpoint_output",
    "mdrun_runtime_args",
]

_named_features_0[3] = [
    "cli_env_kwarg",
]

_named_features_0[4] = ["mpi_comm_integration"]


def api_is_at_least(major_version, minor_version=0, patch_version=0):
    """Allow client to check whether installed module supports the requested API level.

    Arguments:
        major_version (int): gmxapi major version number.
        minor_version (int): optional gmxapi minor version number (default: 0).
        patch_version (int): optional gmxapi patch level number (default: 0).

    Returns:
        True if installed gmx package is greater than or equal to the input level

    Note that if gmxapi.version.release is False, the package is not guaranteed to correctly or
    fully support the reported API level.
    """
    if (
        not isinstance(major_version, int)
        or not isinstance(minor_version, int)
        or not isinstance(patch_version, int)
    ):
        raise TypeError("Version levels must be provided as integers.")
    if _major > major_version:
        return True
    elif _major == major_version and _minor >= minor_version:
        return True
    elif (
        _major == major_version and _minor == minor_version and _micro >= patch_version
    ):
        return True
    else:
        return False


def has_feature(name: str, enable_exception=False) -> bool:
    """Query whether a named feature is available in the installed package.

    Between updates to the API specification, new features or experimental aspects
    may be introduced into the package and need to be detectable. This function
    is intended to facilitate code testing and resolving differences between
    development branches. Users should refer to the documentation for the package
    modules and API level.

    The primary use case is, in conjunction with `api_is_at_least()`, to allow
    client code to robustly identify expected behavior and API support through
    conditional execution and branching. Note that behavior is strongly
    specified by the API major version number. Features that have become part of
    the specification and bug-fixes referring to previous major versions should
    not be checked with *has_feature()*. Using *has_feature()* with old feature
    names will produce a DeprecationWarning for at least one major version, and
    client code should be updated to avoid logic errors in future versions.

    For convenience, setting ``enable_exception = True`` causes the function to
    instead raise a gmxapi.exceptions.FeatureNotAvailableError for unrecognized feature names.
    This allows extension code to cleanly produce a gmxapi exception instead of
    first performing a boolean check. Also, some code may be unexecutable for
    more than one reason, and sometimes it is cleaner to catch all
    `gmxapi.exceptions.Error` exceptions for a code block, rather than to
    construct complex conditionals.

    Returns:
        True if named feature is recognized by the installed package, else False.

    Raises:
        gmxapi.exceptions.FeatureNotAvailableError: If ``enable_exception == True`` and feature is not found.

    """
    import gmxapi._gmxapi as core

    # First, issue a warning if the feature name is subject to removal because
    # of the history of the API specification.
    for version in range(_minor):
        # For sufficiently advanced API versions, we want to warn that old
        # feature checks lose meaning and should no longer be checked.
        # We provide a suggestion with the API version that absorbed their
        # specification.
        if name in _named_features_0[version]:
            warnings.warn(
                f"Old feature name. Use `api_is_at_least({_major}, {version + 1})` instead of `has_feature({name})`.",
                category=DeprecationWarning,
                stacklevel=2,
            )

    # Check whether the feature is listed in the API specification amendments.
    if any(name in features for features in _named_features_0):
        return True
    elif core.has_feature(name):
        return True
    else:
        if enable_exception:
            raise FeatureNotAvailableError(
                "Feature {} not available.".format(str(name))
            )
        return False
